//# {{
#ifdef __INTELLISENSE__
#include "preprocessor.hpp"
#endif
//# }}
TOML_PUSH_WARNINGS;
#ifdef _MSC_VER
#ifndef __clang__
#pragma inline_recursion(on)
#endif
#pragma push_macro("min")
#pragma push_macro("max")
#undef min
#undef max
#endif
